using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using Microsoft.Web.Services3;
using System.Xml.Schema;
using System.Collections;
using System.Diagnostics;

namespace InputValidator
{
    // Validator is part of the Verification component.
    // It communicates with PolicyMappingManager, RepositorySchemaProvider,
    // WsdlSchemaProvider, ValidationLog, CorrectnessVerificator,
    // and MesssageInterceptor.
    // Its task is to validate an incoming SOAP envelope, provided by 
    // MessageInterceptor. To accomplish this, PolicyMappingManager is
    // asked for the input policy for the given web service, the
    // schema providers are asked for the XML schemas, and the envelope
    // is validated against the schemas according to the input policy.
    // The result is then returned to MessageInterceptor.
    public class Validator
    {
        private PolicyMappingManager mappingManager;
        private ISchemaProvider repositorySchemaProvider;
        private WsdlSchemaProvider wsdlSchemaProvider;
        private bool isValid;
        private XmlNamespaceManager xmlnsm;
        private ValidationLog logger;
        

        public Validator(PolicyMappingManager pmm,
          ISchemaProvider repProvider,
          ISchemaProvider wsdlProvider)
        {
            mappingManager = pmm;
            repositorySchemaProvider = repProvider;
            wsdlSchemaProvider = (WsdlSchemaProvider)wsdlProvider;
            logger = new ValidationLog("Validator",
              "InputValidator");
        }

        // The Validator's main interface which returns whether
        // a given SOAP envelope is valid according to the
        // input validation security policy.
        public bool IsValid(SoapEnvelope envelope)
        {
            // Retrieve URI and port from SOAP envelope
            string[] uriAndPort = GetUriAndPort(envelope);
            string uri = uriAndPort[0];
            string port = uriAndPort[1];
            string namesp = uriAndPort[2];

            // Create XML document consisting of
            //only the body
            XmlDocument body = ExtractBody(envelope);

            // Retrieve mapping for this web method
            XmlDocument mapping = 
              mappingManager.GetMapping(uri, port, namesp);

            // Creating namespacemanager
            xmlnsm = 
              new XmlNamespaceManager(mapping.NameTable);
            xmlnsm.AddNamespace("item",
              "http://www.item.ntnu.no/ws");

            // Terminate if validation is not requested
            if (!NeedsValidation(mapping)) return true;      

            if (WsdlMethodValidation(mapping))
            {
                // Validate against WSDL schema

                // Get WSDL schema
                XmlSchema wsdlSchema = 
                  wsdlSchemaProvider.GetSchema(uri);

                // Create XML document consisting of
                //the body contents
                XmlDocument bodyContents = 
                  ExtractBodyContents(body);

                // Validate XML document against WSDL schema
                // If validation fails, return false
                if (!Validate(wsdlSchema, bodyContents))
                {
                    logger.Write("Message did not validate against WSDL schema.\n" + body.InnerXml);
                    return false;
                }

            }
            else
            {
                // Validate against mapping

                // Check that the required parameters, and
                // only the required parameters, are present
                if (!HasPreciselyTheRequiredParameters
                  (mapping, body, port, namesp))
                {
                    logger.Write("Message did not have the required parameters.\n"+  body.InnerXml);
                    return false;
                }

                // Create a list of the parameters that
                // need repository validation
                List<string> repositoryParams = 
                  GetRepositoryParameters(mapping);

                // Get repository schema
                XmlSchema repositorySchema = 
                  repositorySchemaProvider.GetSchema();

                foreach (string param in repositoryParams)
                {
                    // Create an XML document containing
                    // the parameter and value based on
                    // the parameter
                    XmlDocument repositoryParameterDocument =
                      GetRepositoryParameterDocument(
                      param, body, mapping, port, namesp);

                    // Validate XML document against
                    // repository schema
                    // If validation fails, return false
                    if (!Validate(repositorySchema,
                      repositoryParameterDocument))
                    {
                        logger.Write("Message parameter did validate.\n" + repositoryParameterDocument.InnerXml);
                        return false;
                    }
                }
            }

            return true;
        }
        
        private XmlDocument ExtractBodyContents(
          XmlDocument body)
        {
            XmlDocument d = new XmlDocument();
            XmlNode bodyContentNode = 
              body.FirstChild.FirstChild;

            bodyContentNode = (XmlElement)d.ImportNode
              (bodyContentNode, true);
            d.AppendChild(bodyContentNode);
            return d;
        }

        private XmlDocument GetRepositoryParameterDocument
          (string param,
          XmlDocument body,
          XmlDocument mapping,
          string portName,
          string namesp)
        {
            // Get name of schema id from mapping
            XmlNode paramMappingNode = 
              mapping.DocumentElement.SelectSingleNode(
              "./item:param[@name='" + param + "']",
              xmlnsm);
            string schemaId =
              paramMappingNode.Attributes["schemaId"].Value;

            // Get value of parameter from body
            xmlnsm.AddNamespace("ns", namesp);
            XmlNode paramBodyNode = 
              body.DocumentElement.SelectSingleNode(
              "./ns:" + portName + "/ns:" + param, xmlnsm);
            string value = paramBodyNode.FirstChild.Value;

            // Create document containing only an element
            // named schema id and value parameter
            XmlDocument repositoryParameterDocument =
              new XmlDocument();
            XmlElement rootElement = 
              repositoryParameterDocument.CreateElement
                  (schemaId);
            XmlText valueNode = 
              repositoryParameterDocument.CreateTextNode
                  (value);
            rootElement.AppendChild(valueNode);
            repositoryParameterDocument.AppendChild
              (rootElement);

            return repositoryParameterDocument;
        }

        private XmlDocument ExtractBody(
          SoapEnvelope envelope)
        {
            XmlDocument body = new XmlDocument();
            body.Load(envelope.GetBodyReader());
            return body;
        }

        private List<string> GetRepositoryParameters
          (XmlDocument mapping)
        {
            XmlNodeList parameterNodes = 
              mapping.DocumentElement.ChildNodes;
            List<string> parameterNames = new List<string>();
            foreach (XmlNode node in parameterNodes)
            {
                // Skip the parameter if validate is
                // set to false
                if (node.Attributes["validate"] != null 
                  && node.Attributes["validate"].Value 
                  == "false")
                    continue;

                // Skip the parameter if useWsdl is set
                // to true
                if (node.Attributes["useWsdl"] != null
                  && node.Attributes["useWsdl"].Value
                  == "true")
                    continue;

                // Throw an exception if schemaId has not
                // been specified
                if (node.Attributes["schemaId"] == null)
                {
                    logger.Write("schemaId has not been specified for parameter '" + node.Attributes["name"] + "' in port '" + mapping.DocumentElement.LocalName + "'.");
                    throw new Exception("schemaId has not been specified for parameter '" + node.Attributes["name"] + "' in port '" + mapping.DocumentElement.LocalName + "'.");
                }

                // Everything is ok; add the parameter
                // name to the list
                parameterNames.Add(node.Attributes["name"].
                  Value);
            }

            return parameterNames;
        }

        public bool HasPreciselyTheRequiredParameters
          (XmlDocument mapping,
          XmlDocument body,
          string port,
          string namesp)
        {
            // Check that the body tag has one element
            // only, and that this element has the port name
            if (body.DocumentElement.ChildNodes.Count != 1)
            {
                logger.Write("A SOAP envelope with more than one body element was encountered. \n" + body.OuterXml);
                throw new Exception("The body tag can only have one element");
            }

            if (!(body.DocumentElement.FirstChild.LocalName.
              Equals(port) && body.DocumentElement.
              FirstChild.NamespaceURI.Equals(namesp)))
            {
                logger.Write("The SOAP body contents did not have the expected port name and namespace. \n" + body.OuterXml);
                throw new Exception("The tag inside the body element must have the port name and namespace");
            }
            
            // Create a list of required parameters
            // Create a list of parameters present in the
            // SOAP envelope
            // Check that there is a 1-to-1 mapping between
            // the two lists

            XmlNodeList bodyNodeList =
              body.DocumentElement.FirstChild.ChildNodes;
            XmlNodeList mappingList =
              mapping.DocumentElement.ChildNodes;
            List<string> mappingStringList =
              new List<string>();

            // Add the names of all the elements
            foreach (XmlNode node in mappingList)
            {
                if (node.NodeType == XmlNodeType.Element)
                    mappingStringList.Add(
                      node.Attributes["name"].Value);
            }

            
            bool isValid = true;

            if (bodyNodeList.Count ==
              mappingStringList.Count)
              //then begin further parameter check
            {

                foreach (XmlNode node in bodyNodeList)
                {

                    if (mappingStringList.Contains(
                      node.LocalName))
                    {
                        mappingStringList.Remove(
                          node.LocalName);
                    }
                    else
                    {
                        isValid = false;
                        break;
                    }
                }
            }
            else
              // bodynodelist and mappinglist differ
              // on at least one parameter
            {
                isValid = false;
            }

            return isValid;
        }

        
        // Validates an XML document against a given XML Schema
        private bool Validate(
          XmlSchema schema,
          XmlDocument doc)
        {
            isValid = true;
            doc.Schemas.Add(schema);
            ValidationEventHandler eventHandler =
              new ValidationEventHandler(ValidationHandler);
            try
            {
                doc.Validate(eventHandler);

            }
            catch (ArgumentException argE)
            {
                isValid = false;
            }
            catch (XmlSchemaValidationException schemaE)
            {
                isValid = false;
            }
            catch (InvalidOperationException operE)
            {
                isValid = false;

            }

            return isValid;
        }   

        private bool NeedsValidation(XmlDocument mapping)
        {
            XmlNode portNode = mapping.DocumentElement;

            if (portNode.Attributes["validate"] != null 
              && portNode.Attributes["validate"].Value.
              Equals("false"))
            {
                // No validation
                return false;
            }

            return true;
        }

        private bool WsdlMethodValidation(
          XmlDocument mapping)
        {
            XmlNode portNode = mapping.DocumentElement;

            if (portNode.Attributes["useWsdl"] != null &&
              portNode.Attributes["useWsdl"].Value.
              Equals("true"))
            {
                // WSDL method validation
                return true;
            }

            return false;
        }

        private string[] GetUriAndPort(SoapEnvelope envelope)
        {
            string[] uriAndPort = new string[3];

            XmlReader headerReader =
              envelope.GetHeaderReader();
            while (headerReader.Read())
            {
                // Read URI
                if (headerReader.NodeType
                  == XmlNodeType.Element 
                  && headerReader.LocalName 
                  == "To"
                  && headerReader.NamespaceURI
                  == "http://schemas.xmlsoap.org/ws/2004/08/addressing")
                {
                    headerReader.Read();
                    headerReader.MoveToContent();
                    if (headerReader.HasValue)
                    {
                        uriAndPort[0] = headerReader.Value;
                        break;
                    }
                }
            }
            headerReader.Close();

            XmlReader bodyReader = envelope.GetBodyReader();
            bodyReader.Read();
            bodyReader.MoveToContent();
            if (bodyReader.NodeType ==
              XmlNodeType.Element
              && bodyReader.Depth == 1)
            {
                uriAndPort[1] = bodyReader.LocalName;
                uriAndPort[2] = bodyReader.NamespaceURI;
            }
            headerReader.Close();

            return uriAndPort;
        }
        
        // A handler that receives control in the event of
        // validation errors.
        public void ValidationHandler(
          object sender,
          ValidationEventArgs e)
        {
            logger.Write("Validation error:\n" + e.Message);
            isValid = false;
        }
    }
}
